import re
from sqlite3 import connect

import datacompy
import numpy as np
import pandas as pd
import pytest
import tomli_w
import xugrid
from pydantic import ValidationError
from pyproj import CRS
from ribasim import Node
from ribasim.config import Solver
from ribasim.geometry.link import NodeData
from ribasim.input_base import esc_id
from ribasim.model import Model
from ribasim.nodes import basin
from ribasim_testmodels import (
    basic_model,
    outlet_model,
    pid_control_equation_model,
    trivial_model,
)
from shapely import Point


def test_repr(basic):
    representation = repr(basic).split("\n")
    assert representation[0] == "ribasim.Model("


def test_solver():
    solver = Solver()
    assert solver.algorithm == "QNDF"  # default
    assert solver.saveat == 86400.0

    solver = Solver(saveat=3600.0)
    assert solver.saveat == 3600.0

    solver = Solver(saveat=float("inf"))
    assert solver.saveat == float("inf")

    solver = Solver(saveat=0)
    assert solver.saveat == 0

    with pytest.raises(ValidationError):
        Solver(saveat="a")


@pytest.mark.xfail(reason="Needs refactor")
def test_invalid_node_type(basic):
    # Add entry with invalid node type
    basic.node.static = basic.node.df._append(
        {"node_type": "InvalidNodeType", "geometry": Point(0, 0)}, ignore_index=True
    )

    with pytest.raises(
        TypeError,
        match=re.escape("Invalid node types detected: [InvalidNodeType].") + ".+",
    ):
        basic.validate_model_node_types()


def test_parent_relationship(basic):
    model = basic
    assert model.pump._parent == model
    assert model.pump._parent_field == "pump"


def test_exclude_unset(basic):
    model = basic
    model.solver.saveat = 86400.0
    d = model.model_dump(exclude_unset=True, exclude_none=True, by_alias=True)
    assert "solver" in d
    assert d["solver"]["saveat"] == 86400.0


def test_invalid_node_id():
    with pytest.raises(
        ValueError,
        match=r".* Input should be greater than or equal to 0 .*",
    ):
        Node(-1, Point(7.0, 7.0))


def test_tabulated_rating_curve_model(tabulated_rating_curve, tmp_path):
    model_orig = tabulated_rating_curve
    basin_area = tabulated_rating_curve.basin.area.df
    assert basin_area is not None
    assert basin_area.crs == CRS.from_epsg(28992)
    model_orig.set_crs(model_orig.crs)
    assert basin_area.geometry.geom_type.iloc[0] == "MultiPolygon"
    assert basin_area.crs == CRS.from_epsg(28992)
    model_orig.write(tmp_path / "tabulated_rating_curve/ribasim.toml")
    model_new = Model.read(tmp_path / "tabulated_rating_curve/ribasim.toml")
    pd.testing.assert_series_equal(
        model_orig.tabulated_rating_curve.time.df.time,
        model_new.tabulated_rating_curve.time.df.time,
    )


def test_plot(discrete_control_of_pid_control):
    discrete_control_of_pid_control.plot()


def test_write_adds_fid_in_tables(basic, tmp_path):
    model_orig = basic
    # for node an explicit index was provided
    nrow = len(model_orig.basin.node.df)
    assert model_orig.basin.node.df.index.name == "node_id"

    # for link an explicit index was provided
    nrow = len(model_orig.link.df)
    assert model_orig.link.df.index.name == "link_id"
    assert model_orig.link.df.index.equals(pd.RangeIndex(1, nrow + 1))

    # Index name is applied by _name_index
    df = model_orig.link.df.copy()
    df.index.name = "other"
    model_orig.link.df = df
    assert model_orig.link.df.index.name == "link_id"

    model_orig.write(tmp_path / "basic/ribasim.toml")
    with connect(tmp_path / "basic/database.gpkg") as connection:
        query = f"select * from {esc_id('Basin / profile')}"
        df = pd.read_sql_query(query, connection, dtype_backend="pyarrow")
        assert "fid" in df.columns

        query = "select node_id from Node"
        df = pd.read_sql_query(query, connection, dtype_backend="pyarrow")
        assert "node_id" in df.columns

        query = "select link_id from Link"
        df = pd.read_sql_query(query, connection, dtype_backend="pyarrow")
        assert "link_id" in df.columns


def test_node_table(basic):
    model = basic
    assert model.flow_boundary.node.df.crs == CRS.from_epsg(28992)
    node = model.node_table()
    df = node.df
    assert df.geometry.is_unique
    assert df.index.dtype == np.int32
    assert df.subnetwork_id.dtype == pd.Int32Dtype()
    assert df.node_type.iloc[0] == "Basin"
    assert df.node_type.iloc[-1] == "LevelBoundary"
    assert df.crs == CRS.from_epsg(28992)


def test_link_table(basic):
    model = basic
    df = model.link.df
    assert df.geometry.is_unique
    assert df.from_node_id.dtype == np.int32
    assert df.crs == CRS.from_epsg(28992)


def test_indexing(basic):
    model = basic

    result = model.basin[1]
    assert isinstance(result, NodeData)

    # Also test with a numpy type
    result = model.basin[np.int32(1)]
    assert isinstance(result, NodeData)

    with pytest.raises(TypeError, match="Basin index must be an integer, not list"):
        model.basin[[1, 3, 6]]

    result = model.basin.static[1]
    assert isinstance(result, pd.DataFrame)

    result = model.basin.static[[1, 3, 6]]
    assert isinstance(result, pd.DataFrame)

    with pytest.raises(
        IndexError, match=re.escape("Basin / static does not contain node_id: [2]")
    ):
        model.basin.static[2]

    with pytest.raises(
        ValueError,
        match=re.escape("Cannot index into Basin / time: it contains no data."),
    ):
        model.basin.time[1]


@pytest.mark.parametrize(
    "model",
    [basic_model(), outlet_model(), pid_control_equation_model(), trivial_model()],
)
def test_xugrid(model, tmp_path):
    uds = model.to_xugrid(add_flow=False)
    assert isinstance(uds, xugrid.UgridDataset)
    assert uds.grid.edge_dimension == "ribasim_nEdges"
    assert uds.grid.node_dimension == "ribasim_nNodes"
    assert uds.grid.crs == CRS.from_epsg(28992)
    assert uds.node_id.dtype == np.int32
    uds.ugrid.to_netcdf(tmp_path / "ribasim.nc")
    uds = xugrid.open_dataset(tmp_path / "ribasim.nc")
    assert uds.attrs["Conventions"] == "CF-1.9 UGRID-1.0"

    with pytest.raises(FileNotFoundError, match="Model must be written to disk"):
        model.to_xugrid(add_flow=True)

    model.write(tmp_path / "ribasim.toml")
    with pytest.raises(FileNotFoundError, match="Cannot find results"):
        model.to_xugrid(add_flow=True)
    with pytest.raises(FileNotFoundError, match="or allocation is not used"):
        model.to_xugrid(add_flow=False, add_allocation=True)
    with pytest.raises(ValueError, match="Cannot add both allocation and flow results"):
        model.to_xugrid(add_flow=True, add_allocation=True)


def test_to_crs(bucket: Model):
    model = bucket

    # Reproject to World Geodetic System 1984
    model.to_crs("EPSG:4326")

    # Assert that the bucket is still at Deltares' headquarter
    assert model.basin.node.df["geometry"].iloc[0].x == pytest.approx(4.38, abs=0.1)
    assert model.basin.node.df["geometry"].iloc[0].y == pytest.approx(51.98, abs=0.1)


def test_styles(tabulated_rating_curve: Model, tmp_path):
    model = tabulated_rating_curve

    model.write(tmp_path / "basic" / "ribasim.toml")
    with connect(tmp_path / "basic" / "database.gpkg") as conn:
        assert conn.execute("SELECT COUNT(*) FROM layer_styles").fetchone()[0] == 3


def test_non_existent_files(tmp_path):
    with pytest.raises(
        FileNotFoundError, match="File 'non_existent_file.toml' does not exist."
    ):
        Model.read("non_existent_file.toml")

    # Create a TOML file without a database.gpkg
    content = {"input_path": str(tmp_path)}
    toml_path = tmp_path / "test.toml"
    with open(toml_path, "wb") as f:
        tomli_w.dump(content, f)

    with pytest.raises(FileNotFoundError, match=r"Database file .* does not exist\."):
        Model.read(toml_path)


def test_model_equals(basic):
    nbasic = basic.model_copy(deep=True)

    assert nbasic.basin.static == basic.basin.static
    assert nbasic.basin == basic.basin
    assert nbasic == basic

    nbasic.solver.saveat = 0
    assert nbasic.solver.saveat != basic.solver.saveat
    assert nbasic.solver != basic.solver
    assert nbasic.basin == basic.basin
    assert nbasic != basic

    nbasic.solver.saveat = basic.solver.saveat
    nbasic.basin.add(
        Node(None, Point(-1.5, -1), name="confluence"),
        [
            basin.Static(precipitation=[4]),
        ],
    )
    assert nbasic.basin.static != basic.basin.static
    assert nbasic.basin != basic.basin
    assert nbasic != basic


def test_model_diff(basic):
    # Create a copy of the model to compare with
    nbasic = basic.model_copy(deep=True)
    x = nbasic.diff(basic)
    assert x is None

    # Test unequal comparisons for both paths
    with pytest.raises(ValueError):
        nbasic.diff(basic.solver)
    with pytest.raises(ValueError):
        basic.basin.static.diff(basic.basin.node)

    # Change the solver settings and compare
    nbasic.solver.saveat = 0
    x = nbasic.diff(basic)
    assert isinstance(x, dict)
    assert "solver" in x
    assert len(x) == 1  # only solver is different
    assert "saveat" in x["solver"]
    assert x["solver"]["saveat"]["self"] == 0
    assert x["solver"]["saveat"]["other"] == 86400.0

    # Add metadata information to the static node
    nbasic.basin.static.df["meta_data"] = 1
    x = nbasic.basin.static.diff(basic.basin.static, ignore_meta=True)
    assert x is None
    x = nbasic.basin.static.diff(basic.basin.static, ignore_meta=False)
    assert isinstance(x, dict)
    assert "diff" in x
    assert isinstance(x["diff"], datacompy.Compare)

    # Reset and add new basin / static node.
    nbasic.basin.static.df = basic.basin.static.df.copy()
    nbasic.solver.saveat = basic.solver.saveat
    nbasic.basin.add(
        Node(None, Point(-1.5, -1), name="confluence"),
        [
            basin.Static(precipitation=[4], meta_data=1),
        ],
    )

    # Test DataFrame difference on TableModel level
    x = nbasic.basin.static.diff(basic.basin.static)
    assert isinstance(x, dict)
    assert "diff" in x
    assert isinstance(x["diff"], datacompy.Compare)

    # Test DataFrame difference on model level
    x = nbasic.diff(basic)
    assert isinstance(x, dict)
    assert "basin" in x
    assert len(x) == 1  # only basin is different
    assert "static" in x["basin"]
    assert "diff" in x["basin"]["static"]
    assert isinstance(x["basin"]["static"]["diff"], datacompy.Compare)



from ribasim.input_base import TableModel
from ribasim.schemas import (
    FlowBoundaryConcentrationSchema,
    FlowBoundaryStaticSchema,
    FlowBoundaryTimeSchema,
)

__all__ = ["Static", "Time", "Concentration"]


class Static(TableModel[FlowBoundaryStaticSchema]):
    pass


class Time(TableModel[FlowBoundaryTimeSchema]):
    pass


class Concentration(TableModel[FlowBoundaryConcentrationSchema]):
    pass



from .generate import add_tracer, generate
from .parse import parse
from .plot import plot_fraction, plot_spatial
from .util import run_delwaq

__all__ = [
    "generate",
    "parse",
    "run_delwaq",
    "add_tracer",
    "plot_fraction",
    "plot_spatial",
]



import datetime
import logging
import shutil
from collections.abc import Generator
from os import PathLike
from pathlib import Path
from typing import Any

import numpy as np
import pandas as pd
import tomli
import tomli_w
from matplotlib import pyplot as plt
from pandera.typing.geopandas import GeoDataFrame
from pydantic import (
    DirectoryPath,
    Field,
    PrivateAttr,
    field_serializer,
    model_validator,
)

import ribasim
from ribasim.config import (
    Allocation,
    Basin,
    ContinuousControl,
    DiscreteControl,
    Experimental,
    FlowBoundary,
    FlowDemand,
    LevelBoundary,
    LevelDemand,
    LinearResistance,
    Logging,
    ManningResistance,
    MultiNodeModel,
    Outlet,
    PidControl,
    Pump,
    Results,
    Solver,
    TabulatedRatingCurve,
    Terminal,
    UserDemand,
)
from ribasim.db_utils import _set_db_schema_version
from ribasim.geometry.link import LinkSchema, LinkTable
from ribasim.geometry.node import NodeTable
from ribasim.input_base import (
    ChildModel,
    FileModel,
    SpatialTableModel,
    context_file_loading,
    context_file_writing,
)
from ribasim.utils import (
    MissingOptionalModule,
    UsedIDs,
    _concat,
    _link_lookup,
    _node_lookup,
    _node_lookup_numpy,
    _time_in_ns,
)
from ribasim.validation import control_link_neighbor_amount, flow_link_neighbor_amount

try:
    import xugrid
except ImportError:
    xugrid = MissingOptionalModule("xugrid")


class Model(FileModel):
    """A model of inland water resources systems."""

    starttime: datetime.datetime
    endtime: datetime.datetime
    crs: str

    input_dir: Path = Field(default=Path("."))
    results_dir: Path = Field(default=Path("results"))

    logging: Logging = Field(default_factory=Logging)
    solver: Solver = Field(default_factory=Solver)
    results: Results = Field(default_factory=Results)

    allocation: Allocation = Field(default_factory=Allocation)

    experimental: Experimental = Field(default_factory=Experimental)

    basin: Basin = Field(default_factory=Basin)
    continuous_control: ContinuousControl = Field(default_factory=ContinuousControl)
    discrete_control: DiscreteControl = Field(default_factory=DiscreteControl)
    flow_boundary: FlowBoundary = Field(default_factory=FlowBoundary)
    flow_demand: FlowDemand = Field(default_factory=FlowDemand)
    level_boundary: LevelBoundary = Field(default_factory=LevelBoundary)
    level_demand: LevelDemand = Field(default_factory=LevelDemand)
    linear_resistance: LinearResistance = Field(default_factory=LinearResistance)
    manning_resistance: ManningResistance = Field(default_factory=ManningResistance)
    outlet: Outlet = Field(default_factory=Outlet)
    pid_control: PidControl = Field(default_factory=PidControl)
    pump: Pump = Field(default_factory=Pump)
    tabulated_rating_curve: TabulatedRatingCurve = Field(
        default_factory=TabulatedRatingCurve
    )
    terminal: Terminal = Field(default_factory=Terminal)
    user_demand: UserDemand = Field(default_factory=UserDemand)

    link: LinkTable = Field(default_factory=LinkTable)
    use_validation: bool = Field(default=True, exclude=True)

    _used_node_ids: UsedIDs = PrivateAttr(default_factory=UsedIDs)

    @model_validator(mode="after")
    def _set_node_parent(self) -> "Model":
        for (
            k,
            v,
        ) in self._children().items():
            setattr(v, "_parent", self)
            setattr(v, "_parent_field", k)
        return self

    @model_validator(mode="after")
    def _ensure_link_table_is_present(self) -> "Model":
        if self.link.df is None:
            self.link.df = GeoDataFrame[LinkSchema](index=pd.Index([], name="link_id"))
        self.link.df = self.link.df.set_geometry("geometry", crs=self.crs)
        return self

    @model_validator(mode="after")
    def _update_used_ids(self) -> "Model":
        # Only update the used node IDs if we read from a database
        if "database" in context_file_loading.get():
            df = self.node_table().df
            assert df is not None
            if len(df.index) > 0:
                self._used_node_ids.node_ids.update(df.index)
                self._used_node_ids.max_node_id = df.index.max()
        return self

    @field_serializer("input_dir", "results_dir")
    def _serialize_path(self, path: Path) -> str:
        return str(path)

    def model_post_init(self, __context: Any) -> None:
        # When serializing we exclude fields that are set to their default values
        # However, we always want to write `input_dir` and `results_dir`
        # By overriding `BaseModel.model_post_init` we can set them explicitly,
        # and enforce that they are always written.
        self.model_fields_set.update({"input_dir", "results_dir"})
        self.edge = self.link  # Backwards compatible alias for link

    def __repr__(self) -> str:
        """Generate a succinct overview of the Model content.

        Skip "empty" NodeModel instances: when all dataframes are None.
        """
        content = ["ribasim.Model("]
        INDENT = "    "
        for field in self._fields():
            attr = getattr(self, field)
            if isinstance(attr, LinkTable):
                content.append(f"{INDENT}{field}=Link(...),")
            else:
                if isinstance(attr, MultiNodeModel) and attr.node.df is None:
                    # Skip unused node types
                    continue
                content.append(f"{INDENT}{field}={repr(attr)},")

        content.append(")")
        return "\n".join(content)

    def _write_toml(self, fn: Path) -> Path:
        """
        Write the model data to a TOML file.

        Parameters
        ----------
        fn : FilePath
            The file path where the TOML file will be written.

        Returns
        -------
        Path
            The file path of the written TOML file.
        """
        content = self.model_dump(
            exclude_unset=True, exclude_none=True, by_alias=True, context="write"
        )
        # Filter empty dicts (default Nodes)
        content = dict(filter(lambda x: x[1], content.items()))
        content["ribasim_version"] = ribasim.__version__
        with open(fn, "wb") as f:
            tomli_w.dump(content, f)
        return fn

    def _save(self, directory: DirectoryPath, input_dir: DirectoryPath):
        # We write all tables to a temporary GeoPackage with a dot prefix,
        # and at the end move this over the target file.
        # This does not throw a PermissionError if the file is open in QGIS.
        db_path = directory / input_dir / ".database.gpkg"

        # avoid adding tables to existing model
        db_path.parent.mkdir(parents=True, exist_ok=True)
        db_path.unlink(missing_ok=True)
        context_file_writing.get()["database"] = db_path

        self.link._save(directory, input_dir)
        node = self.node_table()

        assert node.df is not None
        node._save(directory, input_dir)

        # Run after geopackage schema has been created
        _set_db_schema_version(db_path, ribasim.__schema_version__)

        for sub in self._nodes():
            sub._save(directory, input_dir)

        shutil.move(db_path, db_path.with_name("database.gpkg"))

    def set_crs(self, crs: str) -> None:
        """Set the coordinate reference system of the data in the model.

        Parameters
        ----------
        crs : str
            Coordinate reference system, like "EPSG:4326" for WGS84 latitude longitude.
        """
        self._apply_crs_function("set_crs", crs)

    def to_crs(self, crs: str) -> None:
        # Set CRS of the tables to the CRS stored in the Model object
        self.set_crs(self.crs)
        self._apply_crs_function("to_crs", crs)

    def _apply_crs_function(self, function_name: str, crs: str) -> None:
        """Apply `function_name`, with `crs` as the first and only argument to all spatial tables."""
        getattr(self.link.df, function_name)(crs, inplace=True)
        for sub in self._nodes():
            if sub.node.df is not None:
                getattr(sub.node.df, function_name)(crs, inplace=True)
            for table in sub._tables():
                if isinstance(table, SpatialTableModel) and table.df is not None:
                    getattr(table.df, function_name)(crs, inplace=True)
        self.crs = crs

    def node_table(self) -> NodeTable:
        """Compute the full sorted NodeTable from all node types."""
        df_chunks = [node.node.df for node in self._nodes()]
        df = (
            _concat(df_chunks)
            if df_chunks
            else pd.DataFrame(index=pd.Index([], name="node_id"))
        )
        node_table = NodeTable(df=df)
        node_table.sort()
        assert node_table.df is not None
        assert node_table.df.index.is_unique, "node_id must be unique"
        return node_table

    def _nodes(self) -> Generator[MultiNodeModel, Any, None]:
        """Return all non-empty MultiNodeModel instances."""
        for key in self.model_fields.keys():
            attr = getattr(self, key)
            if (
                isinstance(attr, MultiNodeModel)
                and attr.node.df is not None
                # TODO: Model.read creates empty node tables (#1278)
                and not attr.node.df.empty
            ):
                yield attr

    def _children(self):
        return {
            k: getattr(self, k)
            for k in self.model_fields.keys()
            if isinstance(getattr(self, k), ChildModel)
        }

    @classmethod
    def read(cls, filepath: str | PathLike[str]) -> "Model":
        """Read a model from a TOML file.

        Parameters
        ----------
        filepath : str | PathLike[str]
            The path to the TOML file.
        """
        if not Path(filepath).is_file():
            raise FileNotFoundError(f"File '{filepath}' does not exist.")
        return cls(filepath=filepath)  # type: ignore

    def write(self, filepath: str | PathLike[str]) -> Path:
        """Write the contents of the model to disk and save it as a TOML configuration file.

        If ``filepath.parent`` does not exist, it is created before writing.

        Parameters
        ----------
        filepath : str | PathLike[str]
            A file path with .toml extension.
        """
        if self.use_validation:
            self._validate_model()

        filepath = Path(filepath)
        self.filepath = filepath
        if not filepath.suffix == ".toml":
            raise ValueError(f"Filepath '{filepath}' is not a .toml file.")
        context_file_writing.set({})
        directory = filepath.parent
        directory.mkdir(parents=True, exist_ok=True)
        self._save(directory, self.input_dir)
        fn = self._write_toml(filepath)

        context_file_writing.set({})
        return fn

    def _validate_model(self):
        df_link = self.link.df
        df_chunks = [node.node.df for node in self._nodes()]
        df_node = _concat(df_chunks)

        df_graph = df_link
        # Join df_link with df_node to get to_node_type
        df_graph = df_graph.join(
            df_node[["node_type"]], on="from_node_id", how="left", rsuffix="_from"
        )
        df_graph = df_graph.rename(columns={"node_type": "from_node_type"})

        df_graph = df_graph.join(
            df_node[["node_type"]], on="to_node_id", how="left", rsuffix="_to"
        )
        df_graph = df_graph.rename(columns={"node_type": "to_node_type"})

        if not self._has_valid_neighbor_amount(
            df_graph, flow_link_neighbor_amount, "flow", df_node["node_type"]
        ):
            raise ValueError("Minimum flow inneighbor or outneighbor unsatisfied")
        if not self._has_valid_neighbor_amount(
            df_graph, control_link_neighbor_amount, "control", df_node["node_type"]
        ):
            raise ValueError("Minimum control inneighbor or outneighbor unsatisfied")

    def _has_valid_neighbor_amount(
        self,
        df_graph: pd.DataFrame,
        link_amount: dict[str, list[int]],
        link_type: str,
        nodes,
    ) -> bool:
        """Check if the neighbor amount of the two nodes connected by the given link meet the minimum requirements."""
        is_valid = True

        # filter graph by link type
        df_graph = df_graph.loc[df_graph["link_type"] == link_type]

        # count occurrence of "from_node" which reflects the number of outneighbors
        from_node_count = (
            df_graph.groupby("from_node_id").size().reset_index(name="from_node_count")
        )

        # append from_node_count column to from_node_id and from_node_type
        from_node_info = (
            df_graph[["from_node_id", "from_node_type"]]
            .drop_duplicates()
            .merge(from_node_count, on="from_node_id", how="left")
        )
        from_node_info = from_node_info[
            ["from_node_id", "from_node_count", "from_node_type"]
        ]

        # add the node that is not the upstream of any other nodes
        from_node_info = self._add_source_sink_node(nodes, from_node_info, "from")

        # loop over all the "from_node" and check if they have enough outneighbor
        for _, row in from_node_info.iterrows():
            # from node's outneighbor
            if row["from_node_count"] < link_amount[row["from_node_type"]][2]:
                is_valid = False
                logging.error(
                    f"Node {row['from_node_id']} must have at least {link_amount[row['from_node_type']][2]} outneighbor(s) (got {row['from_node_count']})"
                )

        # count occurrence of "to_node" which reflects the number of inneighbors
        to_node_count = (
            df_graph.groupby("to_node_id").size().reset_index(name="to_node_count")
        )

        # append to_node_count column to result
        to_node_info = (
            df_graph[["to_node_id", "to_node_type"]]
            .drop_duplicates()
            .merge(to_node_count, on="to_node_id", how="left")
        )
        to_node_info = to_node_info[["to_node_id", "to_node_count", "to_node_type"]]

        # add the node that is not the downstream of any other nodes
        to_node_info = self._add_source_sink_node(nodes, to_node_info, "to")

        # loop over all the "to_node" and check if they have enough inneighbor
        for _, row in to_node_info.iterrows():
            if row["to_node_count"] < link_amount[row["to_node_type"]][0]:
                is_valid = False
                logging.error(
                    f"Node {row['to_node_id']} must have at least {link_amount[row['to_node_type']][0]} inneighbor(s) (got {row['to_node_count']})"
                )

        return is_valid

    def _add_source_sink_node(
        self, nodes, node_info: pd.DataFrame, direction: str
    ) -> pd.DataFrame:
        """Loop over node table.

        Add the nodes whose id are missing in the from_node and to_node column in the link table because they are not the upstream or downstrem of other nodes.

        Specify that their occurrence in from_node table or to_node table is 0.
        """
        # loop over nodes, add the one that is not the downstream (from) or upstream (to) of any other nodes
        for index, node in enumerate(nodes):
            if nodes.index[index] not in node_info[f"{direction}_node_id"].to_numpy():
                new_row = {
                    f"{direction}_node_id": nodes.index[index],
                    f"{direction}_node_count": 0,
                    f"{direction}_node_type": node,
                }
                node_info = _concat(
                    [node_info, pd.DataFrame([new_row])], ignore_index=True
                )

        return node_info

    @classmethod
    def _load(cls, filepath: Path | None) -> dict[str, Any]:
        context_file_loading.set({})

        if filepath is not None and filepath.is_file():
            with open(filepath, "rb") as f:
                config = tomli.load(f)

            directory = filepath.parent / config.get("input_dir", ".")
            context_file_loading.get()["directory"] = directory
            db_path = directory / "database.gpkg"

            if not db_path.is_file():
                raise FileNotFoundError(f"Database file '{db_path}' does not exist.")

            context_file_loading.get()["database"] = db_path

            return config
        else:
            return {}

    @model_validator(mode="after")
    def _reset_contextvar(self) -> "Model":
        # Drop database info
        context_file_loading.set({})
        return self

    def plot_control_listen(self, ax):
        """Plot the implicit listen links of the model."""
        df_listen_link = pd.DataFrame(
            data={
                "control_node_id": pd.Series([], dtype="int32[pyarrow]"),
                "listen_node_id": pd.Series([], dtype="int32[pyarrow]"),
            }
        )

        # Listen links from PidControl
        for table in (self.pid_control.static.df, self.pid_control.time.df):
            if table is None:
                continue

            to_add = table[["node_id", "listen_node_id"]].drop_duplicates()
            to_add.columns = ["control_node_id", "listen_node_id"]
            df_listen_link = _concat([df_listen_link, to_add])

        # Listen links from ContinuousControl and DiscreteControl
        for table, name in (
            (self.continuous_control.variable.df, "ContinuousControl"),
            (self.discrete_control.variable.df, "DiscreteControl"),
        ):
            if table is None:
                continue

            to_add = table[["node_id", "listen_node_id"]].drop_duplicates()
            to_add.columns = [
                "control_node_id",
                "listen_node_id",
            ]
            df_listen_link = _concat([df_listen_link, to_add])

        # Collect geometry data
        node = self.node_table().df
        control_nodes_geometry = df_listen_link.merge(
            node,
            left_on=["control_node_id"],
            right_on=["node_id"],
            how="left",
        )["geometry"]

        listen_nodes_geometry = df_listen_link.merge(
            node,
            left_on=["listen_node_id"],
            right_on=["node_id"],
            how="left",
        )["geometry"]

        # Plot listen links
        for i, (point_listen, point_control) in enumerate(
            zip(listen_nodes_geometry, control_nodes_geometry)
        ):
            ax.plot(
                [point_listen.x, point_control.x],
                [point_listen.y, point_control.y],
                color="gray",
                ls="--",
                label="Listen link" if i == 0 else None,
            )
        return

    def plot(
        self,
        ax=None,
        indicate_subnetworks: bool = True,
        aspect_ratio_bound: float = 0.33,
    ) -> Any:
        """Plot the nodes, links and allocation networks of the model.

        Parameters
        ----------
        ax : matplotlib.pyplot.Artist
            Axes on which to draw the plot.
        indicate_subnetworks : bool
            Whether to indicate subnetworks with a convex hull backdrop.
        aspect_ratio_bound : float
            The maximal aspect ratio in (0,1). The smaller this number, the further the figure
            shape is allowed to be from a square

        Returns
        -------
        ax : matplotlib.pyplot.Artist
            Axis on which the plot is drawn.
        """
        if ax is None:
            _, ax = plt.subplots()
            ax.axis("off")

        node = self.node_table()
        self.link.plot(ax=ax, zorder=2)
        self.plot_control_listen(ax)
        node.plot(ax=ax, zorder=3)

        handles, labels = ax.get_legend_handles_labels()

        if indicate_subnetworks:
            (
                handles_subnetworks,
                labels_subnetworks,
            ) = node.plot_allocation_networks(ax=ax, zorder=1)
            handles += handles_subnetworks
            labels += labels_subnetworks

        ax.legend(handles, labels, loc="lower left", bbox_to_anchor=(1, 0.5))

        # Enforce aspect ratio bound
        xlim = ax.get_xlim()
        ylim = ax.get_ylim()
        xsize = xlim[1] - xlim[0]
        ysize = ylim[1] - ylim[0]

        if ysize < aspect_ratio_bound * xsize:
            y_mid = (ylim[0] + ylim[1]) / 2
            ysize_new = aspect_ratio_bound * xsize
            ax.set_ylim(y_mid - ysize_new / 2, y_mid + ysize_new / 2)
        elif xsize < aspect_ratio_bound * ysize:
            x_mid = (xlim[0] + xlim[1]) / 2
            xsize_new = aspect_ratio_bound * ysize
            ax.set_xlim(x_mid - xsize_new / 2, x_mid + xsize_new / 2)

        return ax

    def to_xugrid(self, add_flow: bool = False, add_allocation: bool = False):
        """Convert the network to a `xugrid.UgridDataset`.

        Either the flow or the allocation data can be added, but not both simultaneously.
        This method will throw `ImportError` if the optional dependency `xugrid` isn't installed.

        Parameters
        ----------
        add_flow : bool
            add flow results (Optional, defaults to False)
        add_allocation : bool
            add allocation results (Optional, defaults to False)
        """
        if add_flow and add_allocation:
            raise ValueError("Cannot add both allocation and flow results.")

        node_df = self.node_table().df
        assert node_df is not None

        assert self.link.df is not None
        link_df = self.link.df.copy()
        # We assume only the flow network is of interest.
        link_df = link_df[link_df.link_type == "flow"]

        node_id = node_df.index.to_numpy()
        link_id = link_df.index.to_numpy()
        from_node_id = link_df.from_node_id.to_numpy()
        to_node_id = link_df.to_node_id.to_numpy()
        node_lookup = _node_lookup_numpy(node_id)

        grid = xugrid.Ugrid1d(
            node_x=node_df.geometry.x,
            node_y=node_df.geometry.y,
            fill_value=-1,
            edge_node_connectivity=np.column_stack(
                (
                    node_lookup.loc[from_node_id],
                    node_lookup.loc[to_node_id],
                )
            ),
            name="ribasim",
            projected=node_df.crs.is_projected,
            crs=node_df.crs,
        )

        link_dim = grid.edge_dimension
        node_dim = grid.node_dimension

        uds = xugrid.UgridDataset(None, grid)
        uds = uds.assign_coords(node_id=(node_dim, node_id))
        uds = uds.assign_coords(link_id=(link_dim, link_id))
        uds = uds.assign_coords(from_node_id=(link_dim, from_node_id))
        uds = uds.assign_coords(to_node_id=(link_dim, to_node_id))

        if add_flow:
            uds = self._add_flow(uds, node_lookup)
        elif add_allocation:
            uds = self._add_allocation(uds)

        return uds

    def _checked_toml_path(self) -> Path:
        toml_path = self.filepath
        if toml_path is None:
            raise FileNotFoundError("Model must be written to disk to add results.")
        return toml_path

    def _add_flow(self, uds, node_lookup):
        toml_path = self._checked_toml_path()

        results_path = toml_path.parent / self.results_dir
        basin_path = results_path / "basin.arrow"
        flow_path = results_path / "flow.arrow"

        if not basin_path.is_file() or not flow_path.is_file():
            raise FileNotFoundError(
                f"Cannot find results in '{results_path}', "
                "perhaps the model needs to be run first."
            )

        basin_df = pd.read_feather(basin_path, dtype_backend="pyarrow")
        flow_df = pd.read_feather(flow_path, dtype_backend="pyarrow")
        _time_in_ns(basin_df)
        _time_in_ns(flow_df)

        # add the xugrid dimension indices to the dataframes
        link_dim = uds.grid.edge_dimension
        node_dim = uds.grid.node_dimension
        node_lookup = _node_lookup(uds)
        link_lookup = _link_lookup(uds)
        flow_df[link_dim] = link_lookup[flow_df["link_id"]].to_numpy()
        basin_df[node_dim] = node_lookup[basin_df["node_id"]].to_numpy()

        # add flow results to the UgridDataset
        flow_da = flow_df.set_index(["time", link_dim])["flow_rate"].to_xarray()
        uds[flow_da.name] = flow_da

        # add basin results to the UgridDataset
        basin_df.drop(columns=["node_id"], inplace=True)
        basin_ds = basin_df.set_index(["time", node_dim]).to_xarray()

        for var_name, da in basin_ds.data_vars.items():
            uds[var_name] = da

        return uds

    def _add_allocation(self, uds):
        toml_path = self._checked_toml_path()

        results_path = toml_path.parent / self.results_dir
        alloc_flow_path = results_path / "allocation_flow.arrow"

        if not alloc_flow_path.is_file():
            raise FileNotFoundError(
                f"Cannot find '{alloc_flow_path}', "
                "perhaps the model needs to be run first, or allocation is not used."
            )

        alloc_flow_df = pd.read_feather(
            alloc_flow_path,
            columns=[
                "time",
                "link_id",
                "flow_rate",
                "optimization_type",
                "demand_priority",
            ],
            dtype_backend="pyarrow",
        )
        _time_in_ns(alloc_flow_df)

        # add the xugrid link dimension index to the dataframe
        link_dim = uds.grid.edge_dimension
        link_lookup = _link_lookup(uds)
        alloc_flow_df[link_dim] = link_lookup[alloc_flow_df["link_id"]].to_numpy()

        # "flow_rate_allocated" is the sum of all allocated flow rates over the demand priorities
        allocate_df = alloc_flow_df.loc[
            alloc_flow_df["optimization_type"] == "allocate"
        ]
        uds["flow_rate_allocated"] = (
            allocate_df.groupby(["time", link_dim])["flow_rate"].sum().to_xarray()
        )

        # also add the individual demand priorities and optimization types
        # added as separate variables to ensure QGIS / MDAL compatibility
        for (optimization_type, demand_priority), group in alloc_flow_df.groupby(
            ["optimization_type", "demand_priority"]
        ):
            varname = f"{optimization_type}_priority_{demand_priority}"
            da = group.set_index(["time", link_dim])["flow_rate"].to_xarray()
            uds[varname] = da

        return uds



from ribasim.input_base import TableModel
from ribasim.schemas import (
    ContinuousControlFunctionSchema,
    ContinuousControlVariableSchema,
)

__all__ = ["Variable", "Function"]


class Variable(TableModel[ContinuousControlVariableSchema]):
    pass


class Function(TableModel[ContinuousControlFunctionSchema]):
    pass



# Automatically generated file. Do not modify.

# Table for connectivity
# "Basin": ["LinearResistance"] means that the downstream of basin can be LinearResistance only
node_type_connectivity: dict[str, list[str]] = {
    "PidControl": [
        "Outlet",
        "Pump",
    ],
    "LevelBoundary": [
        "Outlet",
        "TabulatedRatingCurve",
        "LinearResistance",
        "Pump",
    ],
    "Pump": [
        "LevelBoundary",
        "Basin",
        "Terminal",
    ],
    "UserDemand": [
        "LevelBoundary",
        "Basin",
        "Terminal",
    ],
    "TabulatedRatingCurve": [
        "LevelBoundary",
        "Basin",
        "Terminal",
    ],
    "FlowDemand": [
        "LinearResistance",
        "Outlet",
        "TabulatedRatingCurve",
        "ManningResistance",
        "Pump",
    ],
    "FlowBoundary": [
        "LevelBoundary",
        "Basin",
        "Terminal",
    ],
    "Basin": [
        "LinearResistance",
        "UserDemand",
        "Outlet",
        "TabulatedRatingCurve",
        "ManningResistance",
        "Pump",
    ],
    "ManningResistance": [
        "Basin",
    ],
    "LevelDemand": [
        "Basin",
    ],
    "DiscreteControl": [
        "LinearResistance",
        "PidControl",
        "Outlet",
        "TabulatedRatingCurve",
        "ManningResistance",
        "Pump",
    ],
    "Outlet": [
        "LevelBoundary",
        "Basin",
        "Terminal",
    ],
    "ContinuousControl": [
        "Outlet",
        "Pump",
    ],
    "LinearResistance": [
        "LevelBoundary",
        "Basin",
    ],
    "Terminal": [],
}


# Function to validate connection
def can_connect(node_type_up: str, node_type_down: str) -> bool:
    if node_type_up in node_type_connectivity:
        return node_type_down in node_type_connectivity[node_type_up]
    return False


flow_link_neighbor_amount: dict[str, list[int]] = {
    "PidControl": [0, 0, 0, 0],
    "LevelBoundary": [0, 9223372036854775807, 0, 9223372036854775807],
    "Pump": [1, 1, 1, 1],
    "UserDemand": [1, 1, 1, 1],
    "TabulatedRatingCurve": [1, 1, 1, 1],
    "FlowDemand": [0, 0, 0, 0],
    "FlowBoundary": [0, 0, 1, 9223372036854775807],
    "Basin": [0, 9223372036854775807, 0, 9223372036854775807],
    "ManningResistance": [1, 1, 1, 1],
    "LevelDemand": [0, 0, 0, 0],
    "DiscreteControl": [0, 0, 0, 0],
    "Outlet": [1, 1, 1, 1],
    "ContinuousControl": [0, 0, 0, 0],
    "LinearResistance": [1, 1, 1, 1],
    "Terminal": [1, 9223372036854775807, 0, 0],
}

control_link_neighbor_amount: dict[str, list[int]] = {
    "PidControl": [0, 1, 1, 1],
    "LevelBoundary": [0, 0, 0, 0],
    "Pump": [0, 1, 0, 0],
    "UserDemand": [0, 0, 0, 0],
    "TabulatedRatingCurve": [0, 1, 0, 0],
    "FlowDemand": [0, 0, 1, 1],
    "FlowBoundary": [0, 0, 0, 0],
    "Basin": [0, 1, 0, 0],
    "ManningResistance": [0, 1, 0, 0],
    "LevelDemand": [0, 0, 1, 9223372036854775807],
    "DiscreteControl": [0, 0, 1, 9223372036854775807],
    "Outlet": [0, 1, 0, 0],
    "ContinuousControl": [0, 0, 1, 9223372036854775807],
    "LinearResistance": [0, 1, 0, 0],
    "Terminal": [0, 0, 0, 0],
}



import logging
from contextlib import closing
from datetime import datetime
from pathlib import Path
from sqlite3 import connect

STYLES_DIR = Path(__file__).parent / "styles"

CREATE_TABLE_SQL = """
CREATE TABLE IF NOT EXISTS "main"."layer_styles" (
    "id" INTEGER PRIMARY KEY AUTOINCREMENT NOT NULL,
    "f_table_catalog" TEXT(256),
    "f_table_schema" TEXT(256),
    "f_table_name" TEXT(256),
    "f_geometry_column" TEXT(256),
    "styleName" TEXT(30),
    "styleQML" TEXT,
    "styleSLD" TEXT,
    "useAsDefault" BOOLEAN,
    "description" TEXT,
    "owner" TEXT(30),
    "ui" TEXT(30),
    "update_time" DATETIME DEFAULT (strftime('%Y-%m-%dT%H:%M:%fZ','now'))
);
"""

INSERT_CONTENTS_SQL = """
INSERT INTO gpkg_contents (
    "table_name",
    "data_type",
    "identifier",
    "description",
    "last_change",
    "min_x",
    "min_y",
    "max_x",
    "max_y",
    "srs_id"
)
VALUES (
    'layer_styles',
    'attributes',
    'layer_styles',
    '',
    '',
    NULL,
    NULL,
    NULL,
    NULL,
    0
);
"""

INSERT_ROW_SQL = """
INSERT INTO "main"."layer_styles" (
    "f_table_catalog",
    "f_table_schema",
    "f_table_name",
    "f_geometry_column",
    "styleName",
    "styleQML",
    "styleSLD",
    "useAsDefault",
    "description",
    "owner",
    "ui",
    "update_time"
)
VALUES (
    '',
    '',
    :layer,
    'geom',
    :style_name,
    :style_qml,
    '',
    '1',
    :description,
    '',
    NULL,
    :update_date_time
);
"""

SQL_STYLES_EXIST = """
SELECT EXISTS(SELECT 1 FROM sqlite_master WHERE type='table' AND name='layer_styles');
"""

SQL_STYLES_STYLE_EXISTS = """
SELECT EXISTS(SELECT 1 FROM layer_styles WHERE styleName = :style_name);
"""


def _no_existing_style(conn, style_name):
    style_exists = conn.execute(
        SQL_STYLES_STYLE_EXISTS, {"style_name": style_name}
    ).fetchone()[0]
    return not style_exists


def _add_styles_to_geopackage(gpkg_path: Path, layer: str):
    with closing(connect(gpkg_path)) as connection:
        if not connection.execute(SQL_STYLES_EXIST).fetchone()[0]:
            connection.execute(CREATE_TABLE_SQL)
            connection.execute(INSERT_CONTENTS_SQL)

        style_name = f"{layer.replace(' / ', '_')}Style"
        style_qml = STYLES_DIR / f"{style_name}.qml"

        if style_qml.exists() and _no_existing_style(connection, style_name):
            description = f"Ribasim style for layer: {layer}"
            update_date_time = f"{datetime.now().isoformat()}Z"

            connection.execute(
                INSERT_ROW_SQL,
                {
                    "layer": layer,
                    "style_qml": style_qml.read_bytes(),
                    "style_name": style_name,
                    "description": description,
                    "update_date_time": update_date_time,
                },
            )
            connection.commit()
        else:
            logging.warning(f"Style not found for layer: {layer}")



from ribasim.input_base import TableModel
from ribasim.schemas import (
    UserDemandConcentrationSchema,
    UserDemandStaticSchema,
    UserDemandTimeSchema,
)

__all__ = ["Static", "Time"]


class Static(TableModel[UserDemandStaticSchema]):
    pass


class Time(TableModel[UserDemandTimeSchema]):
    pass


class Concentration(TableModel[UserDemandConcentrationSchema]):
    pass



from ribasim.input_base import TableModel
from ribasim.schemas import PidControlStaticSchema, PidControlTimeSchema

__all__ = ["Static", "Time"]


class Static(TableModel[PidControlStaticSchema]):
    pass


class Time(TableModel[PidControlTimeSchema]):
    pass



from ribasim.input_base import TableModel
from ribasim.schemas import (
    LevelBoundaryConcentrationSchema,
    LevelBoundaryStaticSchema,
    LevelBoundaryTimeSchema,
)

__all__ = ["Static", "Time", "Concentration"]


class Static(TableModel[LevelBoundaryStaticSchema]):
    pass


class Time(TableModel[LevelBoundaryTimeSchema]):
    pass


class Concentration(TableModel[LevelBoundaryConcentrationSchema]):
    pass



import re
from warnings import catch_warnings, filterwarnings

import numpy as np
import pandas as pd
from pandera.dtypes import Int32
from pandera.typing import Series
from pydantic import BaseModel, NonNegativeInt


def _pascal_to_snake(pascal_str):
    # Insert a '_' before all uppercase letters that are not at the start of the string
    # and convert the string to lowercase
    return re.sub(r"(?<!^)(?=[A-Z])", "_", pascal_str).lower()


class MissingOptionalModule:
    """Presents a clear error for optional modules."""

    def __init__(self, name, suggestion="all"):
        self.name = name
        self.suggestion = suggestion

    def __getattr__(
        self,
        _,
    ):
        raise ImportError(
            f"{self.name} is required for this functionality. You can get it using `pip install ribasim[{self.suggestion}]`."
        )


def _node_lookup_numpy(node_id) -> Series[Int32]:
    """Create a lookup table from from node_id to the node dimension index.

    Used when adding data onto the nodes of an xugrid dataset.
    """
    return pd.Series(
        index=node_id,
        data=node_id.argsort().astype(np.int32),
        name="node_index",
    )


def _node_lookup(uds) -> Series[Int32]:
    """Create a lookup table from from node_id to the node dimension index.

    Used when adding data onto the nodes of an xugrid dataset.
    """
    return pd.Series(
        index=uds["node_id"],
        data=uds[uds.grid.node_dimension],
        name="node_index",
    )


def _link_lookup(uds) -> Series[Int32]:
    """Create a lookup table from link_id to the link dimension index.

    Used when adding data onto the links of an xugrid dataset.
    """
    return pd.Series(
        index=uds["link_id"],
        data=uds[uds.grid.edge_dimension],
        name="link_index",
    )


def _time_in_ns(df) -> None:
    """Convert the time column to datetime64[ns] dtype."""
    # datetime64[ms] gives trouble; https://github.com/pydata/xarray/issues/6318
    df["time"] = df["time"].astype("datetime64[ns]")


def _concat(dfs, **kwargs):
    """Concatenate DataFrames with a warning filter."""
    with catch_warnings():
        # The behavior of array concatenation with empty entries is deprecated.
        # In a future version, this will no longer exclude empty items when determining
        # the result dtype. To retain the old behavior, exclude the empty entries before
        # the concat operation.
        filterwarnings(
            "ignore",
            category=FutureWarning,
        )
        return pd.concat(dfs, **kwargs)


class UsedIDs(BaseModel):
    """A helper class to manage globally unique node IDs.

    We keep track of all IDs in the model,
    and keep track of the maximum to provide new IDs.
    MultiNodeModels and Link will check this instance on `add`.
    """

    node_ids: set[int] = set()
    max_node_id: NonNegativeInt = 0

    def add(self, node_id: int) -> None:
        self.node_ids.add(node_id)
        self.max_node_id = max(self.max_node_id, node_id)

    def __contains__(self, value: int) -> bool:
        return self.node_ids.__contains__(value)

    def new_id(self) -> int:
        return self.max_node_id + 1



from ribasim.input_base import TableModel
from ribasim.schemas import OutletStaticSchema, OutletTimeSchema

__all__ = ["Static", "Time"]


class Static(TableModel[OutletStaticSchema]):
    pass


class Time(TableModel[OutletTimeSchema]):
    pass



from typing import Any, get_type_hints

import pandera as pa
from geopandas import GeoSeries as _GeoSeries
from pandera.typing import Series
from pandera.typing.geopandas import GeoSeries

from ribasim.schemas import _BaseSchema


class _GeoBaseSchema(_BaseSchema):
    @pa.check("geometry")
    def is_correct_geometry_type(cls, geoseries: GeoSeries[Any]) -> Series[bool]:
        T = get_type_hints(cls)["geometry"].__args__[0]
        return geoseries.map(lambda geom: isinstance(geom, T))

    @pa.parser("geometry")
    def force_2d(cls, geometry: GeoSeries[Any]) -> GeoSeries[Any]:
        if isinstance(geometry, _GeoSeries):
            # Pandera requires GeoSeries to have a name
            return GeoSeries(geometry.force_2d(), name=geometry.name)
        else:
            return geometry



import warnings

from geopandas import GeoDataFrame
from pandas import DataFrame

# On each breaking change, increment the __schema_version__ by one.
# Do the same for write_schema_version in ribasim_qgis/core/geopackage.py


def _rename_column(df, from_colname, to_colname):
    """Rename a column, ensuring we don't end up with two of the same name."""
    # If a column has a default value, or is nullable, they are always added.
    # Remove that column first, then rename the old column.
    if to_colname in df.columns and from_colname not in df.columns:
        warnings.warn(
            "Already migrated, your model (version) might be inconsistent.", UserWarning
        )
        return df

    df.drop(columns=to_colname, inplace=True, errors="ignore")
    df.rename(columns={from_colname: to_colname}, inplace=True, errors="raise")
    return df


def nodeschema_migration(gdf: GeoDataFrame, schema_version: int) -> GeoDataFrame:
    if schema_version == 0 and "node_id" in gdf.columns:
        warnings.warn("Migrating outdated Node table.", UserWarning)
        assert gdf["node_id"].is_unique, "Node IDs have to be unique."
        gdf.set_index("node_id", inplace=True)

    return gdf


def linkschema_migration(gdf: GeoDataFrame, schema_version: int) -> GeoDataFrame:
    if schema_version == 0:
        warnings.warn("Migrating outdated Link table.", UserWarning)
        gdf.drop(columns="from_node_type", inplace=True, errors="ignore")
    if schema_version == 0:
        warnings.warn("Migrating outdated Link table.", UserWarning)
        gdf.drop(columns="to_node_type", inplace=True, errors="ignore")
    if "edge_id" in gdf.columns and schema_version == 0:
        warnings.warn("Migrating outdated Link table.", UserWarning)
        assert gdf["edge_id"].is_unique, "Link IDs have to be unique."
        gdf.set_index("edge_id", inplace=True)
    if schema_version < 3 and "subnetwork_id" in gdf.columns:
        warnings.warn("Migrating outdated Link table.", UserWarning)
        gdf.drop(columns="subnetwork_id", inplace=True, errors="ignore")
    if schema_version < 4 and gdf.index.name == "edge_id":
        warnings.warn("Migrating outdated Link table.", UserWarning)
        gdf.index.rename("link_id", inplace=True)
    if schema_version < 4 and "edge_type" in gdf.columns:
        warnings.warn("Migrating outdated Link table.", UserWarning)
        _rename_column(gdf, "edge_type", "link_type")

    return gdf


def basinstaticschema_migration(df: DataFrame, schema_version: int) -> DataFrame:
    if schema_version == 0:
        warnings.warn("Migrating outdated Basin / static table.", UserWarning)
        df.drop(columns="urban_runoff", inplace=True, errors="ignore")

    return df


def basintimeschema_migration(df: DataFrame, schema_version: int) -> DataFrame:
    if schema_version == 0:
        warnings.warn("Migrating outdated Basin / time table.", UserWarning)
        df.drop(columns="urban_runoff", inplace=True, errors="ignore")

    return df


def continuouscontrolvariableschema_migration(
    df: DataFrame, schema_version: int
) -> DataFrame:
    if schema_version == 0:
        warnings.warn(
            "Migrating outdated ContinuousControl / variable table.", UserWarning
        )
        df.drop(columns="listen_node_type", inplace=True, errors="ignore")

    return df


def discretecontrolvariableschema_migration(
    df: DataFrame, schema_version: int
) -> DataFrame:
    if schema_version == 0:
        warnings.warn(
            "Migrating outdated DiscreteControl / variable table.", UserWarning
        )
        df.drop(columns="listen_node_type", inplace=True, errors="ignore")

    return df


def pidcontrolstaticschema_migration(df: DataFrame, schema_version: int) -> DataFrame:
    if schema_version == 0:
        warnings.warn("Migrating outdated PidControl / static table.", UserWarning)
        df.drop(columns="listen_node_type", inplace=True, errors="ignore")

    return df


def outletstaticschema_migration(df: DataFrame, schema_version: int) -> DataFrame:
    if schema_version < 2:
        warnings.warn("Migrating outdated Outlet / static table.", UserWarning)
        _rename_column(df, "min_crest_level", "min_upstream_level")

    return df


for node_type in ["UserDemand", "LevelDemand", "FlowDemand"]:
    for table_type in ["static", "time"]:

        def migration_func(
            df: DataFrame,
            schema_version: int,
            node_type: str = node_type,
            table_type: str = table_type,
        ) -> DataFrame:
            if schema_version < 4:
                warnings.warn(
                    f"Migrating outdated {node_type} / {table_type} table.", UserWarning
                )
                df.rename(columns={"priority": "demand_priority"}, inplace=True)
            return df

        globals()[f"{node_type.lower()}{table_type}schema_migration"] = migration_func


def discretecontrolconditionschema_migration(
    df: DataFrame, schema_version: int
) -> DataFrame:
    if schema_version < 5:
        warnings.warn(
            "Migrating outdated DiscreteControl / condition table.", UserWarning
        )
        n_rows = len(df)
        df["time"] = [None] * n_rows
        df["condition_id"] = range(1, n_rows + 1)
    return df
